#
# Copyright (C) 2007 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

__all__ = ['derive_record_format',
           'parse_audit_text_input',
           'get_audit_dict',
           
           'AuditEventID',
           'AuditEvent',
           'AuditRecord',
           'AuditRecordReader',
           
           ]

import audit
import copy
import os
import re
import struct

from setroubleshoot.log import *

#-----------------------------------------------------------------------------

# FIXME, this is temporary
def get_audit_dict(event):
    audit_dict = {}
    for record in event.records:
        audit_dict.update(record.fields)
        if record.record_type in [audit.AUDIT_AVC, audit.AUDIT_USER_AVC]:
            audit_dict['access'] = record.get_field('seperms')

    return audit_dict

def derive_record_format(socket_path):
    if re.search('/audispd_events$', socket_path):
        return AuditRecordReader.TEXT_FORMAT
    if re.search('/audit_events$', socket_path):
        return AuditRecordReader.BINARY_FORMAT
    return AuditRecordReader.TEXT_FORMAT                       # assume new format


# regular expression to find message like this:
# msg=audit(1152828325.857:123085): avc:  denied  { append } for  pid=14205 ...
# Note, messages arriving directly from the audit system omit
# 'msg=', but messages in log files prepend 'msg='
# group 1  is the optional "host=XXX "
# group 2  is the host if host=XXX is present
# group 3  is the optional "type=XXX "
# group 4  is the type if type=XXX is present
# group 5  is the optional 'msg='
# group 6  is the complete event id
# group 7  is the seconds component of the timestamp
# group 8  is the millisconds component of the timestamp
# group 9  is the timestamp unique number
# group 10 is the body of the message appearing after the event id
audit_input_re = re.compile('(host=(\S+)\s+)?(type=(\S+)\s+)?(msg=)?audit\(((\d+)\.(\d+):(\d+))\):\s*(.*)')


def parse_audit_text_input(input):
    parse_succeeded = False
    host = None
    record_type = None
    event_id = None
    body_text = None

    match = audit_input_re.search(input)
    if match is not None:
        parse_succeeded = True

        if match.group(2):
            host = match.group(2)

        if match.group(4):
            record_type = audit.audit_name_to_msg_type(match.group(4))

        if match.group(6):
            seconds = int(match.group(7))
            milli   = int(match.group(8))
            serial  = int(match.group(9))
            event_id = AuditEventID(seconds, milli, serial, host)

        body_text = match.group(10)

    return (parse_succeeded, record_type, event_id, body_text)

audit_binary_input_re = re.compile('audit\(((\d+)\.(\d+):(\d+))\):\s*(.*)')

def parse_audit_binary_text(input):
    parse_succeeded = False
    event_id = None
    body_text = None

    match = audit_binary_input_re.search(input)
    if match is not None:
        parse_succeeded = True

        if match.group(1):
            seconds = int(match.group(2))
            milli   = int(match.group(3))
            serial  = int(match.group(4))
            event_id = AuditEventID(seconds, milli, serial)

        body_text = match.group(5)

    return (parse_succeeded, event_id, body_text)

#-----------------------------------------------------------------------------

class AuditEventID:
    def __init__(self, seconds=None, milli=None, serial=None, host=None):
        self.seconds = seconds
        self.milli   = milli
        self.serial  = serial
        self.host    = host

    def __eq__(self, other):
        if self.host    != other.host:          return False
        if self.seconds != other.seconds:       return False
        if self.milli   != other.milli:         return False
        if self.serial  != other.serial:        return False
        return True

    def __cmp__(self, other):
        if self.host != other.host:
            raise ValueError("cannot compare two %s objects whose host values differ (%s!=%s)" \
                             % (self.__class__.__name__, self.host, other.host))

        result = cmp(self.seconds, other.seconds)
        if result != 0: return result

        result = cmp(self.milli, other.milli)
        if result != 0: return result

        result = cmp(self.serial, other.serial)
        if result != 0: return result

        return 0
    
    def copy(self):
        return copy.copy(self)

    time = property(lambda self: float(self.sec) + self.milli / 1000.0)

    def __str__(self):
        return "audit(%d.%d:%d)" % (self.seconds, self.milli, self.serial)

    def is_valid(self):
        if self.seconds is None: return False
        if self.milli   is None: return False
        if self.serial  is None: return False
        return True
    
#-----------------------------------------------------------------------------

class AuditRecord:

    binary_version = 0
    binary_header_format="iiii"
    binary_header_size = struct.calcsize(binary_header_format)
    key_value_pair_re = re.compile("([^ \t]+)\s*=\s*([^ \t]+)")
    avc_re = re.compile("avc:\s+([^\s]+)\s+{([^}]+)}\s+for\s+")

    def __init__(self, record_type, event_id, body_text, fields, line_number=None):
        # Header
        self.record_type = record_type
        self.event_id = event_id
        self.body_text = body_text
        if isinstance(fields, dict):
            self.fields = fields
        else:
            self.set_fields_from_text(body_text)
        self.line_number = line_number
        self.interpret()
        
    def __str__(self):
        return self.to_text()

    def is_valid(self):
        if not self.event_id.is_valid(): return False
        if self.record_type is None:     return False
        if self.message is None:         return False
        return True

    def set_fields_from_text(self, body_text):
        self.fields = {}

        for match in AuditRecord.key_value_pair_re.finditer(body_text):
            key   = match.group(1)
            value = match.group(2)
            value = value.strip('"')
            self.fields[key] = value

    def get_field(self, name):
        return self.fields.get(name)

    def interpret(self):
        if self.record_type in [audit.AUDIT_AVC, audit.AUDIT_USER_AVC]:
            if not self.fields.has_key('seresult'):
                match = AuditRecord.avc_re.search(self.body_text)
                if match:
                    seresult = match.group(1)
                    self.fields['seresult'] = seresult

                    seperms = match.group(2)
                    self.fields['seperms'] = seperms.split()

        return True

    def get_binary_header(self, msg):
        msg_length = len(msg)
        return struct.pack(AuditRecord.binary_header_format, AuditRecord.binary_version,
                           AuditRecord.binary_header_size, self.record_type, msg_length)

    def fields_to_text(self):
        if self.fields is None: return ''
        keys = self.fields.keys()
        keys.sort()
        return ' '.join(["%s=%s" % (k, self.fields[k]) for k in keys])

    def to_text(self):
        return "type=%s msg=%s: %s\n" % \
               (audit.audit_msg_type_to_name(self.record_type), self.event_id, self.body_text)

    def to_host_text(self):
        if self.event_id.host is not None:
            return "host=%s type=%s msg=%s: %s\n" % \
                   (self.event_id.host, audit.audit_msg_type_to_name(self.record_type),
                    self.event_id, self.body_text)
        else:
            return self.to_text()

    def to_binary(self):
        record = "%s: %s" % (self.event_id, self.body_text)
        return self.get_binary_header(record) + record

#-----------------------------------------------------------------------------

class AuditRecordReader:
    BINARY_FORMAT = 1
    TEXT_FORMAT = 2

    def __init__(self, record_format):
        self.record_format = record_format
        self._input_buffer = ''
        self.line_number = 0
        
        if self.record_format == self.TEXT_FORMAT:
            self.feed = self.feed_text
        elif self.record_format == self.BINARY_FORMAT:
            self.feed = self.feed_binary
        else:
            raise ValueError("unknown record format (%s) in %s" % (record_format, self.__class__.__name__))

    def feed_binary(self, new_data):
        if len(new_data) <= 0:
            return
        self._input_buffer += new_data

        # Now process as much of the buffer as we can, iterating over complete
        # messages.

        while True:

            # To read a complete message there must be a complete header and
            # all the data the header specified via the header.length
            if len(self._input_buffer) < AuditRecord.binary_header_size:
                return

            binary_version, binary_header_size, record_type, msg_length = \
                            struct.unpack(AuditRecord.binary_header_format,
                                          self._input_buffer[0:AuditRecord.binary_header_size])

            total_len = AuditRecord.binary_header_size + msg_length

            if len(self._input_buffer) < total_len:
                return

            text = self._input_buffer[AuditRecord.binary_header_size:total_len]
            parse_succeeded, event_id, body_text = parse_audit_binary_text(text)
            self._input_buffer = self._input_buffer[total_len:]

            if parse_succeeded:
                yield (record_type, event_id, body_text, None, 0)

        return

    def feed_text(self, new_data):
        if len(new_data) <= 0:
            return
        self._input_buffer += new_data

        # Now process as much of the buffer as we can, iterating over complete
        # messages.

        # To read a complete message we must see a line ending
        start = 0
        end = self._input_buffer.find('\n', start)
        while end >= 0:
            self.line_number += 1
            end += 1                # include newline
            line = self._input_buffer[start:end]
            parse_succeeded, record_type, event_id, body_text = parse_audit_text_input(line)
            if parse_succeeded:
                yield (record_type, event_id, body_text, None, self.line_number)
            start = end
            end = self._input_buffer.find('\n', start)

        self._input_buffer = self._input_buffer[start:]

        return

#-----------------------------------------------------------------------------

class AuditEvent:
    def __init__(self):
        self.event_id     = None
        self.records      = []
        self.record_types = {}

    def __str__(self):
        line_numbers = self.line_numbers
        line_numbers.sort()
        return "%s: is_avc=%s, is_granted=%s: line_numbers=[%s]\n%s" % \
               (self.event_id, self.is_avc(), self.is_granted(),
                ",".join([str(x) for x in line_numbers]),
                "\n".join(["    %s" % record for record in self.records]))

    line_numbers = property(lambda self: [record.line_number for record in self.records if record.line_number])

    def add_record(self, record):
        if self.event_id is None:
            self.event_id = record.event_id.copy()
            self.timestamp = float(self.event_id.seconds) + (self.event_id.milli / 1000.0)
        else:
            if not self.event_id == record.event_id:
                raise ValueError("cannot add audit record to audit event, event_id mismatch %s != %s" % \
                                 (self.event_id, record.event_id))

        self.records.append(record)
        record_list = self.record_types.setdefault(record.record_type, [])
        record_list.append(record)

    def get_field(self, name, record_type=None):
        '''Return list of (value, record_type) tuples.
        In other words return the value matching name for every record_type.
        If record_type is not specified then all records are searched.
        Note: it is possible to have more than one record of a given type
        thus it is always possible to have multiple values returned.'''
        items = []
        if record_type is None:
            records = self.records
        else:
            records = self.get_records_of_type(record_type)

        for record in self.records:
            value = record.fields.get(name)
            if value is None: continue
            items.append((value, record.type))

        return items

    def get_records_of_type(self, type):
        return self.record_types.get(type)

    def get_avc_record(self):
        record_list = self.record_types.get(audit.AUDIT_AVC)
        if record_list is None:
            record_list = self.get_records_of_type(audit.AUDIT_USER_AVC)
        if record_list is not None:
            return record_list[0]
        return None

    def is_avc(self):
        return self.get_avc_record() is not None

    def is_granted(self):
        avc_record = self.get_avc_record()
        if avc_record is None:
            return False
        seresult = avc_record.fields['seresult']
        if seresult == 'denied':
            return False
        if seresult == 'granted':
            return True
        log.avc.warn("unknown value for seresult ('%s')", seresult)
        return False

